!
! Copyright (c) 2012-2013,2015-2017 Yoshihide Yoshimoto
!

#include "fftgrid-vx.h"

module exchange
  integer,private:: blk_nw = 20
  real(kind=8),private,parameter:: rcf = 1.0d0, rdf = sqrt(2.0d0)
  real(kind=8),private:: rc = -1.0d0, rd = -1.0d0

  logical,private:: commij_init = .false.
  integer,private:: comm_parent, commi, commj, niproc, njproc
  integer,private:: myirank, myjrank
  integer,private:: gam_avail, gam_stat, nwary(3)

  private:: engexch_subr, clear_tmp, reflesh_commij
contains
  subroutine reflesh_commij(myrank,nproc,comm,mssg)
    use mpi
    implicit none
    integer,intent(in):: myrank, nproc, comm, mssg
    integer:: ierr, retval

    if (commij_init) then
       call MPI_COMM_COMPARE(comm,comm_parent,retval,ierr)
       if (retval .ne. MPI_IDENT) then
          call MPI_COMM_FREE(commi,ierr)
          call MPI_COMM_FREE(commj,ierr)
          commij_init = .false.
          comm_parent = MPI_COMM_NULL
       end if
    end if
    if (.not. commij_init) then
       do niproc=int(sqrt(dble(nproc))),nproc
          if (mod(nproc,niproc).eq.0) exit
       end do
       njproc = nproc/niproc
       if (mssg.gt.0) then
          write(6,'(1x,a,i5,1x,i5)') &
               'potexch: niproc, njproc = ', niproc, njproc
       end if

       call MPI_COMM_SPLIT(comm,myrank/niproc,mod(myrank,niproc),commi,ierr)
       call MPI_COMM_SPLIT(comm,mod(myrank,niproc),myrank/niproc,commj,ierr)
       call MPI_COMM_RANK(commi,myirank,ierr)
       call MPI_COMM_RANK(commj,myjrank,ierr)

       comm_parent = comm
       commij_init = .true.
    end if
    return
  end subroutine reflesh_commij

  subroutine clear_tmp(x,n)
    implicit none
    integer,intent(in):: n
    real(kind=8),intent(out):: x(n)
    integer:: i
#ifdef __FUJITSU
!OCL XFILL
    do i=1,n
       x(i) = 0.0d0
    end do
#else
!$omp parallel do private(i)
    do i=1,n
       x(i) = 0.0d0
    end do
#endif
  end subroutine clear_tmp

  subroutine potexch_setcutoff(xctp,qexat,aa0,nkmesh,mssg)
    use util
    use xcf_types
    implicit none
    real(kind=8),intent(in):: aa0(3,3),qexat(2)
    integer,intent(in):: xctp,nkmesh,mssg

    real(kind=8),parameter:: pi = 3.14159265358979323846d0

    real(kind=8):: va0,r0

    if (xcf_exch_factor(xctp) .eq. 0.0d0) then
       return
    end if

    if (qexat(1) .lt. 0.0d0) then
       if (mssg.gt.0) then
          write(6,*) 'potexch: automatically set rc, rd from va0'
          write(6,'(1x,a,e18.11,1x,e18.11)') 'potexch: rcf, rdf =', rcf,rdf
       end if
       va0 = cellvol(aa0)
       r0 = ( 3.0d0*(nkmesh*va0)/(4.0d0*pi) )**(1.0d0/3.0d0)
       rc = r0*rcf
       rd = r0*rdf
    else
       rc = qexat(1)
       rd = qexat(2)
    end if

    if (mssg.gt.0) then
       write(6,'(1x,a,e18.11,1x,e18.11)') 'potexch: rc, rd =', rc,rd
    end if
    return
  end subroutine potexch_setcutoff

  subroutine potexch_init(ifgam,nwx,nwy,nwz,mssg)
    implicit none
    integer:: ifgam,nwx,nwy,nwz,mssg

    external init_exchange

    gam_avail = ifgam
    nwary(1) = nwx
    nwary(2) = nwy
    nwary(3) = nwz
    gam_stat = ifgam
    call init_exchange(gam_stat,blk_nw,blk_nw,nwx,nwy,nwz)
    !
    ! init_exchange may disable gam_stat
    !
    if (mssg.gt.0) then
       write(6,'(1x,a,i4)') 'potexch: blk_nw = ', blk_nw
    end if

    return
  end subroutine potexch_init

  subroutine potexch_done()
    use mpi
    implicit none
    external teardown_exchange
    integer:: ierr

    if (commij_init) then
       call MPI_COMM_FREE(commi,ierr)
       call MPI_COMM_FREE(commj,ierr)
       commij_init = .false.
       comm_parent = MPI_COMM_NULL
    end if

    call teardown_exchange()
    nwary(1) = 0
    nwary(2) = 0
    nwary(3) = 0
    gam_avail = 0
    gam_stat = 0
    return
  end subroutine potexch_done

  subroutine potexch_reflesh(enable_gam)
    implicit none
    integer:: enable_gam
    external init_exchange
    external teardown_exchange

    if (gam_avail.eq.0) return
    if (gam_avail.eq.1 .and. gam_stat.eq.enable_gam) return

    call teardown_exchange()

    if (enable_gam.eq.1) then
       call init_exchange(1,blk_nw,blk_nw,nwary(1),nwary(2),nwary(3),0)
       gam_stat = 1
    else
       call init_exchange(0,blk_nw,blk_nw,nwary(1),nwary(2),nwary(3),0)
       gam_stat = 0
    end if

    return
  end subroutine potexch_reflesh

  subroutine potexch(va,bb,mgsf,nw,nb1,nb2,wgtk,bzwk,dpf,ski,nkmi,bkmi,ekmi, &
       kgi,dp,dvxp,nproc,g_count,g_displ,comm,mssg)
    use types
    use symwfn
    use timeprof
    use mpi
    use meshgrid
    implicit none

    real(kind=8),intent(in):: va,bb(3,3)
    type (meshgrid_struct):: mgsf
    integer:: nw
    integer,intent(in):: mssg
    type (wfn_for_k):: dpf(mgsf%nks)
    real(kind=8),intent(in):: ski(3)
    integer,intent(in):: nkmi,bkmi,ekmi,kgi(3,nkmi)
    complex(kind=8):: dp(bkmi:ekmi,nw),dvxp(bkmi:ekmi,nw)
    integer,intent(in):: nb1,nb2
    real(kind=8),intent(in):: wgtk(nb1:nb2,mgsf%nks),bzwk(mgsf%nks)
    integer,intent(in):: nproc,g_count(0:nproc-1),g_displ(0:nproc-1),comm

    integer:: ng,nnp,nks
    integer,pointer:: rg(:,:,:),pg(:,:),srg(:)
    integer:: ldkg
    integer,pointer:: nkmary(:),kg(:,:,:)
    integer:: nkmesh,nwx2,nwy2,nwz2

    real(kind=8):: skj(3),trssk(3),sq(3)
    integer,allocatable:: trskg(:,:)
    real(kind=8),allocatable:: vxp(:,:,:,:,:)
    real(kind=8),allocatable:: pxi(:,:,:,:,:),pxj(:,:,:,:,:)
    complex(kind=8):: ctemp
    real(kind=8),allocatable:: wgt(:)
    complex(kind=8),allocatable:: dpfwf(:,:),rbufwf(:)

    integer:: nlx,nly,nlz,nlxyz
    integer:: ix,iy,iz
    integer:: miw, mjw

    real(kind=8),parameter:: pi = 3.1415926535897932385d0

    integer:: iib,ijb,ib,jb,ig,i1,i2,i3,jks,irg,jrg
    real(kind=8):: ph
    integer:: iph
    complex(kind=8),allocatable:: zph(:)
    integer:: nb2eff
    real(kind=8):: epsnb2 = 64.0d0*epsilon(1.0d0)

    complex(kind=8),dimension(:),allocatable:: tdp,tdvxp
    integer,dimension(:),allocatable:: bi_count,bi_displ,bj_count,bj_displ
    integer,dimension(:),allocatable:: s_count,s_displ,r_count,r_displ
    integer:: myrank,irnk,krnk,iig,jg,ic,bjw,ejw

    integer:: ierr

    external setup_exchange_clkern, compute_exchange

    call tpf_begin(tpf_potexch)

    ng = mgsf%ng
    nnp = mgsf%nnp
    nks = mgsf%nks
    rg => mgsf%rg
    pg => mgsf%pg
    srg => mgsf%srg
    ldkg = mgsf%ldkg
    nkmary => mgsf%nkm
    kg => mgsf%kg
    nkmesh = mgsf%nkmesh
    nwx2 = mgsf%nwx*2
    nwy2 = mgsf%nwy*2
    nwz2 = mgsf%nwz*2

    call get_exchange_nlxyz(nlx,nly,nlz)
    nlxyz = nlx*nly*nlz

    call MPI_COMM_RANK(comm,myrank,ierr)
    allocate(zph(-nnp+1:nnp-1),s_count(0:nproc-1),s_displ(0:nproc-1), &
         r_count(0:nproc-1),r_displ(0:nproc-1), stat=ierr)
    call chkalloc('potexch','tmp',ierr)

    call reflesh_commij(myrank,nproc,comm,mssg)

    allocate(bi_count(0:niproc-1),bi_displ(0:niproc-1), &
         bj_count(0:njproc-1),bj_displ(0:njproc-1))

    bi_count = nw/niproc
    do irnk=0,mod(nw,niproc)-1
       bi_count(irnk) = bi_count(irnk) + 1
    end do
    bi_displ(0) = 0
    do irnk=0,niproc-2
       bi_displ(irnk+1) = bi_displ(irnk) + bi_count(irnk)
    end do

    allocate(FFT_A3M_DEFSZ(pxi,nlx,nly,nlz,bi_count(myirank)))
    allocate(tdp(nkmi*bi_count(myirank)),stat=ierr)
    call chkalloc('potexch','tmp2',ierr)

    do krnk=0,nproc-1
       irnk = mod(krnk,niproc)
       s_count(krnk) = g_count(myrank)*bi_count(irnk)
       s_displ(krnk) = g_count(myrank)*bi_displ(irnk)
    end do

    call tpf_begin(tpf_mpi)
    call MPI_ALLTOALL(s_count,1,MPI_INTEGER,r_count,1,MPI_INTEGER,comm,ierr)
    call tpf_end(tpf_mpi)
    r_displ(0) = 0
    do irnk=0,nproc-2
       r_displ(irnk+1) = r_displ(irnk) + r_count(irnk)
    end do

    call tpf_begin(tpf_mpi)
    call MPI_ALLTOALLV(dp,s_count,s_displ,MPI_DOUBLE_COMPLEX, &
         tdp,r_count,r_displ,MPI_DOUBLE_COMPLEX,comm,ierr)
    call tpf_end(tpf_mpi)
    
    call clear_tmp(pxi,nlxyz*2*bi_count(myirank))
!$omp parallel do private(ig,ix,iy,iz,i1,i2,i3,irnk,ib,jg,iig)
    do irnk=0,nproc-1
       do ib=1,bi_count(myirank)
          jg = r_displ(irnk)+g_count(irnk)*(ib-1)
          iig = g_displ(irnk)
!OCL NOVREC(pxi)
!VOCL LOOP,DISJOINT(pxi)
!CDIR NODEP(pxi)
!DEC$ IVDEP
!VOPTION INDEP(pxi)
          do ig=1,g_count(irnk)
             ix = kgi(1,iig+ig)
             iy = kgi(2,iig+ig)
             iz = kgi(3,iig+ig)
             i1=mod(nwx2+ix,nwx2)+1
             i2=mod(nwy2+iy,nwy2)+1
             i3=mod(nwz2+iz,nwz2)+1

             FFT_A3M_RL(pxi,i1,i2,i3,ib) = dble(tdp(jg+ig))
             FFT_A3M_IM(pxi,i1,i2,i3,ib) = aimag(tdp(jg+ig))
          end do
       end do
    end do
    deallocate(tdp)

    allocate(trskg(3,ldkg), stat=ierr)
    allocate(FFT_A3M_DEFSZ(pxj,nlx,nly,nlz,blk_nw), &
         FFT_A3M_DEFSZ(vxp,nlx,nly,nlz,bi_count(myirank)), wgt(blk_nw))

    do iph=-nnp+1,nnp-1
       ph = 2.0d0*pi/dble(nnp)*iph
       zph(iph) = dcmplx( cos(ph),-sin(ph) )
    end do

    call clear_tmp(vxp,nlxyz*2*bi_count(myirank))

    do jks = 1, nks
       do jb=nb2,nb1,-1
          if ( abs(wgtk(jb,jks)) .gt. epsnb2 ) then
             exit
          end if
       end do
       nb2eff = jb

       bj_count = nb2eff/njproc
       do irnk=0,mod(nb2eff,njproc)-1
          bj_count(irnk) = bj_count(irnk) + 1
       end do
       bj_displ(0) = 0
       do irnk=0,njproc-2
          bj_displ(irnk+1) = bj_displ(irnk) + bj_count(irnk)
       end do

       allocate(dpfwf(nkmary(jks),bj_count(myjrank)))
       allocate(rbufwf(nkmary(jks)*bj_count(myjrank)))

       call tpf_begin(tpf_mpi)

       do krnk=0,nproc-1
          irnk = krnk/niproc
          s_count(krnk) = mgsf%g_count(myrank,jks)*bj_count(irnk)
          s_displ(krnk) = mgsf%g_count(myrank,jks)*bj_displ(irnk)
       end do

       call MPI_ALLTOALL(s_count,1,MPI_INTEGER,r_count,1,MPI_INTEGER,comm,ierr)
       r_displ(0) = 0
       do irnk=0,nproc-2
          r_displ(irnk+1) = r_displ(irnk) + r_count(irnk)
       end do
       call MPI_ALLTOALLV(dpf(jks)%wf,s_count,s_displ,MPI_DOUBLE_COMPLEX, &
            rbufwf,r_count,r_displ,MPI_DOUBLE_COMPLEX,comm,ierr)

       call tpf_end(tpf_mpi)

!$omp parallel do private(irnk,ic,jb,jg,ig)
       do irnk=0,nproc-1
          ic = r_displ(irnk)
          jg = mgsf%g_displ(irnk,jks)
          do jb=1,bj_count(myjrank)
             do ig=1,mgsf%g_count(irnk,jks)
                ic = ic + 1
                dpfwf(jg+ig,jb) = rbufwf(ic)
             end do
          end do
       end do

       deallocate(rbufwf)

       skj = mgsf%irdkpt(jks)%sk
       do irg=1,mgsf%irdkpt(jks)%ngir
          jrg = mgsf%irdkpt(jks)%irrg(irg)
          call transskkg(skj,nkmary(jks),kg(:,:,jks), &
               rg(:,:,jrg),srg(jrg),trssk,trskg)

          sq = ski - trssk

          call tpf_begin(tpf_exclkern)
          call setup_exchange_clkern(nkmesh,bb,sq,rc,rd)
          call tpf_end(tpf_exclkern)

          bjw = bj_displ(myjrank)+1
          ejw = bj_displ(myjrank)+bj_count(myjrank)
          do ijb=bjw,ejw,blk_nw
             mjw = min(ijb+blk_nw-1,ejw) - ijb + 1

             do jb=ijb,min(ijb+blk_nw-1,ejw)
                if (nb1.le.jb .and. jb.le.nb2eff) then
                   !
                   ! spin degeneracy is included in bzwk(jks)
                   !
                   wgt(jb-ijb+1) = wgtk(jb,jks)/bzwk(jks)/dble(nkmesh)/va
                else
                   wgt(jb-ijb+1) = 1.0d0/dble(nkmesh)/va
                end if
             end do

             call clear_tmp(pxj,nlxyz*2*blk_nw)
             if (srg(mgsf%irdkpt(jks)%irrg(irg)).eq.1) then
!$omp parallel do private(ig,ctemp,iph,ix,iy,iz,i1,i2,i3)
                do jb=ijb,min(ijb+blk_nw-1,ejw)
                   do ig=1,nkmary(jks)
                      ctemp = dpfwf(ig,jb-bj_displ(myjrank))
                      iph = pg(1,jrg)*trskg(1,ig)+pg(2,jrg)*trskg(2,ig) &
                           +pg(3,jrg)*trskg(3,ig)
                      iph = mod(iph,nnp)
                      ctemp = zph(iph)*ctemp

                      ix = trskg(1,ig)
                      iy = trskg(2,ig)
                      iz = trskg(3,ig)
                      i1=mod(nwx2+ix,nwx2)+1
                      i2=mod(nwy2+iy,nwy2)+1
                      i3=mod(nwz2+iz,nwz2)+1
                      FFT_A3M_RL(pxj,i1,i2,i3,jb-ijb+1) = dble(ctemp)
                      FFT_A3M_IM(pxj,i1,i2,i3,jb-ijb+1) = aimag(ctemp)
                   end do
                end do
             else
!$omp parallel do private(ig,ctemp,iph,ix,iy,iz,i1,i2,i3)
                do jb=ijb,min(ijb+blk_nw-1,ejw)
                   do ig=1,nkmary(jks)
                      ctemp = dpfwf(ig,jb-bj_displ(myjrank))
                      iph = pg(1,jrg)*trskg(1,ig)+pg(2,jrg)*trskg(2,ig) &
                           +pg(3,jrg)*trskg(3,ig)
                      iph = mod(iph,nnp)
                      ctemp = zph(iph)*conjg(ctemp)

                      ix = trskg(1,ig)
                      iy = trskg(2,ig)
                      iz = trskg(3,ig)
                      i1=mod(nwx2+ix,nwx2)+1
                      i2=mod(nwy2+iy,nwy2)+1
                      i3=mod(nwz2+iz,nwz2)+1
                      FFT_A3M_RL(pxj,i1,i2,i3,jb-ijb+1) = dble(ctemp)
                      FFT_A3M_IM(pxj,i1,i2,i3,jb-ijb+1) = aimag(ctemp)
                   end do
                end do
             end if

             !
             ! deploy_pxj_kern() may modify pxj itself
             !
             call deploy_pxj_kern(mjw,pxj)

             do iib=1,bi_count(myirank),blk_nw
                miw = min(iib+blk_nw-1,bi_count(myirank)) - iib + 1

                call tpf_begin(tpf_cexch)
                call compute_exchange(miw,mjw, &
                     FFT_A3M_RL(pxi,1,1,1,iib),pxj, &
                     FFT_A3M_RL(vxp,1,1,1,iib),wgt)
                call tpf_end(tpf_cexch)
             end do
          end do
       end do

       deallocate(dpfwf)
    end do

    deallocate(pxi,trskg,pxj,wgt)

    do krnk=0,nproc-1
       irnk = mod(krnk,niproc)
       s_count(krnk) = g_count(myrank)*bi_count(irnk)
       s_displ(krnk) = g_count(myrank)*bi_displ(irnk)
    end do

    call tpf_begin(tpf_mpi)
    call MPI_ALLTOALL(s_count,1,MPI_INTEGER,r_count,1,MPI_INTEGER,comm,ierr)
    call tpf_end(tpf_mpi)
    r_displ(0) = 0
    do irnk=0,nproc-2
       r_displ(irnk+1) = r_displ(irnk) + r_count(irnk)
    end do

    allocate(tdvxp(nkmi*bi_count(myirank)))

!$omp parallel do private(ig,ix,iy,iz,i1,i2,i3,irnk,ib,jg,iig)
    do irnk=0,nproc-1
       do ib=1,bi_count(myirank)
          jg = r_displ(irnk)+g_count(irnk)*(ib-1)
          iig = g_displ(irnk)
!OCL NOVREC(tdvxp)
!VOCL LOOP,DISJOINT(tdvxp)
!CDIR NODEP(tdvxp)
!DEC$ IVDEP
!VOPTION INDEP(tdvxp)
          do ig=1,g_count(irnk)
             ix = kgi(1,iig+ig)
             iy = kgi(2,iig+ig)
             iz = kgi(3,iig+ig)
             i1=mod(nwx2+ix,nwx2)+1
             i2=mod(nwy2+iy,nwy2)+1
             i3=mod(nwz2+iz,nwz2)+1

             tdvxp(jg+ig) = dcmplx(FFT_A3M_RL(vxp,i1,i2,i3,ib), &
                  FFT_A3M_IM(vxp,i1,i2,i3,ib))
          end do
       end do
    end do

    deallocate(vxp)
    allocate(tdp(nkmi*bi_count(myirank)))

    call tpf_begin(tpf_mpi)
    call MPI_REDUCE(tdvxp,tdp,nkmi*bi_count(myirank), &
         MPI_DOUBLE_COMPLEX,MPI_SUM,0,commj,ierr)
    call tpf_end(tpf_mpi)

    if (myjrank.ne.0) then
       r_count = 0
    end if
    do irnk=0,nproc-1
       if ( (irnk/niproc) .gt. 0) then
          s_count(irnk) = 0
       end if
    end do

    call tpf_begin(tpf_mpi)
    call MPI_ALLTOALLV(tdp,r_count,r_displ,MPI_DOUBLE_COMPLEX, &
         dvxp,s_count,s_displ,MPI_DOUBLE_COMPLEX,comm,ierr)
    call tpf_end(tpf_mpi)

    deallocate(tdp,tdvxp,bi_count,bi_displ,bj_count,bj_displ, &
         s_count,s_displ,r_count,r_displ,zph)

    call tpf_end(tpf_potexch)

    return
  end subroutine potexch

  subroutine potexchg(va,bb,mgsf,nw,nb1,nb2,wgtk,bzwk,dpgf,nkm,bkm,ekm, &
       kg,dp,dvxp,nproc,g_count,g_displ,comm,mssg)
    use symwfn
    use timeprof
    use mpi
    use meshgrid
    implicit none

    real(kind=8),intent(in):: va,bb(3,3)
    type (meshgrid_struct):: mgsf
    integer:: nw
    integer,intent(in):: mssg
    integer,intent(in):: nkm,bkm,ekm,kg(3,nkm)
    complex(kind=8),intent(in):: dpgf(bkm:ekm,nw)
    complex(kind=8):: dp(bkm:ekm,nw),dvxp(bkm:ekm,nw)
    integer,intent(in):: nb1,nb2
    real(kind=8),intent(in):: wgtk(nb1:nb2),bzwk
    integer,intent(in):: nproc,g_count(0:nproc-1),g_displ(0:nproc-1),comm

    integer:: nkmesh,nwx2,nwy2,nwz2

    real(kind=8):: sq(3)
    real(kind=8),allocatable:: vxp(:,:,:,:,:)
    real(kind=8),allocatable:: pxi(:,:,:,:,:),pxj(:,:,:,:,:)
    real(kind=8),allocatable:: wgt(:)
    complex(kind=8),allocatable:: dpfwf(:,:),rbufwf(:)

    integer:: nlx,nly,nlz,nlxyz
    integer:: ix,iy,iz
    integer:: miw, mjw

    real(kind=8),parameter:: pi = 3.1415926535897932385d0

    integer:: iib,ijb,ib,jb,jbk,ig,i1,i2,i3
    integer:: nb2eff
    real(kind=8):: epsnb2 = 64.0d0*epsilon(1.0d0)

    complex(kind=8),dimension(:),allocatable:: tdp,tdvxp
    integer,dimension(:),allocatable:: bi_count,bi_displ,bj_count,bj_displ
    integer,dimension(:),allocatable:: s_count,s_displ,r_count,r_displ
    integer:: myrank,irnk,krnk,iig,jg,ic,bjw,ejw

    integer:: ierr

    external setup_exchange_clkern, compute_exchange

    call tpf_begin(tpf_potexch)

    nkmesh = mgsf%nkmesh
    nwx2 = mgsf%nwx*2
    nwy2 = mgsf%nwy*2
    nwz2 = mgsf%nwz*2

    call get_exchange_nlxyz(nlx,nly,nlz)
    nlxyz = nlx*nly*nlz

    call MPI_COMM_RANK(comm,myrank,ierr)
    allocate(s_count(0:nproc-1),s_displ(0:nproc-1), &
         r_count(0:nproc-1),r_displ(0:nproc-1), stat=ierr)
    call chkalloc('potexchg','tmp',ierr)

    call reflesh_commij(myrank,nproc,comm,mssg)

    allocate(bi_count(0:niproc-1),bi_displ(0:niproc-1), &
         bj_count(0:njproc-1),bj_displ(0:njproc-1))

    bi_count = nw/niproc
    do irnk=0,mod(nw,niproc)-1
       bi_count(irnk) = bi_count(irnk) + 1
    end do
    bi_displ(0) = 0
    do irnk=0,niproc-2
       bi_displ(irnk+1) = bi_displ(irnk) + bi_count(irnk)
    end do

    allocate(FFT_A3M_DEFSZ(pxi,nlx,nly,nlz,bi_count(myirank)))
    allocate(tdp(nkm*bi_count(myirank)),stat=ierr)
    call chkalloc('potexchg','tmp2',ierr)

    do krnk=0,nproc-1
       irnk = mod(krnk,niproc)
       s_count(krnk) = g_count(myrank)*bi_count(irnk)
       s_displ(krnk) = g_count(myrank)*bi_displ(irnk)
    end do

    call tpf_begin(tpf_mpi)
    call MPI_ALLTOALL(s_count,1,MPI_INTEGER,r_count,1,MPI_INTEGER,comm,ierr)
    call tpf_end(tpf_mpi)
    r_displ(0) = 0
    do irnk=0,nproc-2
       r_displ(irnk+1) = r_displ(irnk) + r_count(irnk)
    end do

    call tpf_begin(tpf_mpi)
    call MPI_ALLTOALLV(dp,s_count,s_displ,MPI_DOUBLE_COMPLEX, &
         tdp,r_count,r_displ,MPI_DOUBLE_COMPLEX,comm,ierr)
    call tpf_end(tpf_mpi)
    
    call clear_tmp(pxi,nlxyz*2*bi_count(myirank))
!$omp parallel do private(ig,ix,iy,iz,i1,i2,i3,irnk,ib,jg,iig)
    do irnk=0,nproc-1
       do ib=1,bi_count(myirank)
          jg = r_displ(irnk)+g_count(irnk)*(ib-1)
          iig = g_displ(irnk)
!OCL NOVREC(pxi)
!VOCL LOOP,DISJOINT(pxi)
!CDIR NODEP(pxi)
!DEC$ IVDEP
!VOPTION INDEP(pxi)
          do ig=1,g_count(irnk)
             ix = kg(1,iig+ig)
             iy = kg(2,iig+ig)
             iz = kg(3,iig+ig)
             i1=mod(nwx2+ix,nwx2)+1
             i2=mod(nwy2+iy,nwy2)+1
             i3=mod(nwz2+iz,nwz2)+1

             FFT_A3M_RL(pxi,i1,i2,i3,ib) = dble(tdp(jg+ig))
             FFT_A3M_IM(pxi,i1,i2,i3,ib) = aimag(tdp(jg+ig))
          end do
       end do
    end do
    deallocate(tdp)

!$omp parallel do private(ix,iy,iz,i1,i2,i3)
    do ib=1,bi_count(myirank)
       iz = 1 ! = modulo(0,nwz2)+1
       i3 = 1 ! = modulo(nwz2-0,nwz2)+1
       do iy=2,nwy2/2
          i2=nwy2-iy+2
          FFT_A3M_RL(pxi,1,i2,i3,ib) = FFT_A3M_RL(pxi,1,iy,iz,ib)
          FFT_A3M_IM(pxi,1,i2,i3,ib) = -FFT_A3M_IM(pxi,1,iy,iz,ib)
       end do
       do iz=2,nwz2
          do iy=2,nwy2/2
             i2=nwy2-iy+2
             i3=nwz2-iz+2
             FFT_A3M_RL(pxi,1,i2,i3,ib) = FFT_A3M_RL(pxi,1,iy,iz,ib)
             FFT_A3M_IM(pxi,1,i2,i3,ib) = -FFT_A3M_IM(pxi,1,iy,iz,ib)
          end do
       end do
       do iz=2,nwz2/2
          i3=nwz2-iz+2
          FFT_A3M_RL(pxi,1,1,i3,ib) = FFT_A3M_RL(pxi,1,1,iz,ib)
          FFT_A3M_IM(pxi,1,1,i3,ib) = -FFT_A3M_IM(pxi,1,1,iz,ib)
       end do
    end do

    allocate(FFT_A3M_DEFSZ(pxj,nlx,nly,nlz,blk_nw), &
         FFT_A3M_DEFSZ(vxp,nlx,nly,nlz,bi_count(myirank)), wgt(blk_nw))

    call clear_tmp(vxp,nlxyz*2*bi_count(myirank))

    do jb=nb2,nb1,-1
       if ( abs(wgtk(jb)) .gt. epsnb2 ) then
          exit
       end if
    end do
    nb2eff = jb

    bj_count = nb2eff/njproc
    do irnk=0,mod(nb2eff,njproc)-1
       bj_count(irnk) = bj_count(irnk) + 1
    end do
    bj_displ(0) = 0
    do irnk=0,njproc-2
       bj_displ(irnk+1) = bj_displ(irnk) + bj_count(irnk)
    end do

    allocate(dpfwf(nkm,bj_count(myjrank)))
    allocate(rbufwf(nkm*bj_count(myjrank)))

    call tpf_begin(tpf_mpi)

    do krnk=0,nproc-1
       irnk = krnk/niproc
       s_count(krnk) = g_count(myrank)*bj_count(irnk)
       s_displ(krnk) = g_count(myrank)*bj_displ(irnk)
    end do

    call MPI_ALLTOALL(s_count,1,MPI_INTEGER,r_count,1,MPI_INTEGER,comm,ierr)
    r_displ(0) = 0
    do irnk=0,nproc-2
       r_displ(irnk+1) = r_displ(irnk) + r_count(irnk)
    end do
    call MPI_ALLTOALLV(dpgf,s_count,s_displ,MPI_DOUBLE_COMPLEX, &
         rbufwf,r_count,r_displ,MPI_DOUBLE_COMPLEX,comm,ierr)

    call tpf_end(tpf_mpi)

!$omp parallel do private(irnk,ic,jb,jg,ig)
    do irnk=0,nproc-1
       ic = r_displ(irnk)
       jg = g_displ(irnk)
       do jb=1,bj_count(myjrank)
          do ig=1,g_count(irnk)
             ic = ic + 1
             dpfwf(jg+ig,jb) = rbufwf(ic)
          end do
       end do
    end do

    deallocate(rbufwf)

    sq = 0.0d0

    call tpf_begin(tpf_exclkern)
    call setup_exchange_clkern(nkmesh,bb,sq,rc,rd)
    call tpf_end(tpf_exclkern)

    bjw = bj_displ(myjrank)+1
    ejw = bj_displ(myjrank)+bj_count(myjrank)
    do ijb=bjw,ejw,blk_nw
       mjw = min(ijb+blk_nw-1,ejw) - ijb + 1

       do jb=ijb,min(ijb+blk_nw-1,ejw)
          if (nb1.le.jb .and. jb.le.nb2eff) then
             !
             ! spin degeneracy is included in bzwk(jks)
             !
             wgt(jb-ijb+1) = wgtk(jb)/bzwk/dble(nkmesh)/va
          else
             wgt(jb-ijb+1) = 1.0d0/dble(nkmesh)/va
          end if
       end do

       call clear_tmp(pxj,nlxyz*2*blk_nw)
!$omp parallel do private(ig,ix,iy,iz,i1,i2,i3,jbk)
       do jb=ijb,min(ijb+blk_nw-1,ejw)
          jbk = jb-ijb+1
          do ig=1,nkm
             ix = kg(1,ig)
             iy = kg(2,ig)
             iz = kg(3,ig)
             i1=mod(nwx2+ix,nwx2)+1
             i2=mod(nwy2+iy,nwy2)+1
             i3=mod(nwz2+iz,nwz2)+1
             FFT_A3M_RL(pxj,i1,i2,i3,jbk) = dble(dpfwf(ig,jb-bj_displ(myjrank)))
             FFT_A3M_IM(pxj,i1,i2,i3,jbk) = aimag(dpfwf(ig,jb-bj_displ(myjrank)))
          end do

          iz=1 ! = modulo(0,nwz2)+1
          i3=1 ! = modulo(nwz2-0,nwz2)+1
          do iy=2,nwy2/2
             i2=nwy2-iy+2
             FFT_A3M_RL(pxj,1,i2,i3,jbk) = FFT_A3M_RL(pxj,1,iy,iz,jbk)
             FFT_A3M_IM(pxj,1,i2,i3,jbk) = -FFT_A3M_IM(pxj,1,iy,iz,jbk)
          end do
          do iz=2,nwz2
             do iy=2,nwy2/2
                i2=nwy2-iy+2
                i3=nwz2-iz+2
                FFT_A3M_RL(pxj,1,i2,i3,jbk) = FFT_A3M_RL(pxj,1,iy,iz,jbk)
                FFT_A3M_IM(pxj,1,i2,i3,jbk) = -FFT_A3M_IM(pxj,1,iy,iz,jbk)
             end do
          end do
          do iz=2,nwz2/2
             i3=nwz2-iz+2
             FFT_A3M_RL(pxj,1,1,i3,jbk) = FFT_A3M_RL(pxj,1,1,iz,jbk)
             FFT_A3M_IM(pxj,1,1,i3,jbk) = -FFT_A3M_IM(pxj,1,1,iz,jbk)
          end do
       end do

       !
       ! deploy_pxj_kern() may modify pxj itself
       !
       call deploy_pxj_kern(mjw,pxj)

       do iib=1,bi_count(myirank),blk_nw
          miw = min(iib+blk_nw-1,bi_count(myirank)) - iib + 1

          call tpf_begin(tpf_cexch)
          call compute_exchange(miw,mjw, &
               FFT_A3M_RL(pxi,1,1,1,iib),pxj, &
               FFT_A3M_RL(vxp,1,1,1,iib),wgt)
          call tpf_end(tpf_cexch)
       end do
    end do

    deallocate(dpfwf)

    deallocate(pxi,pxj,wgt)

    do krnk=0,nproc-1
       irnk = mod(krnk,niproc)
       s_count(krnk) = g_count(myrank)*bi_count(irnk)
       s_displ(krnk) = g_count(myrank)*bi_displ(irnk)
    end do

    call tpf_begin(tpf_mpi)
    call MPI_ALLTOALL(s_count,1,MPI_INTEGER,r_count,1,MPI_INTEGER,comm,ierr)
    call tpf_end(tpf_mpi)
    r_displ(0) = 0
    do irnk=0,nproc-2
       r_displ(irnk+1) = r_displ(irnk) + r_count(irnk)
    end do

    allocate(tdvxp(nkm*bi_count(myirank)))

!$omp parallel do private(ig,ix,iy,iz,i1,i2,i3,irnk,ib,jg,iig)
    do irnk=0,nproc-1
       do ib=1,bi_count(myirank)
          jg = r_displ(irnk)+g_count(irnk)*(ib-1)
          iig = g_displ(irnk)
!OCL NOVREC(tdvxp)
!VOCL LOOP,DISJOINT(tdvxp)
!CDIR NODEP(tdvxp)
!DEC$ IVDEP
!VOPTION INDEP(tdvxp)
          do ig=1,g_count(irnk)
             ix = kg(1,iig+ig)
             iy = kg(2,iig+ig)
             iz = kg(3,iig+ig)
             i1=mod(nwx2+ix,nwx2)+1
             i2=mod(nwy2+iy,nwy2)+1
             i3=mod(nwz2+iz,nwz2)+1

             tdvxp(jg+ig) = dcmplx(FFT_A3M_RL(vxp,i1,i2,i3,ib), &
                  FFT_A3M_IM(vxp,i1,i2,i3,ib))
          end do
       end do
    end do

    deallocate(vxp)
    allocate(tdp(nkm*bi_count(myirank)))

    call tpf_begin(tpf_mpi)
    call MPI_REDUCE(tdvxp,tdp,nkm*bi_count(myirank), &
         MPI_DOUBLE_COMPLEX,MPI_SUM,0,commj,ierr)
    call tpf_end(tpf_mpi)

    if (myjrank.ne.0) then
       r_count = 0
    end if
    do irnk=0,nproc-1
       if ( (irnk/niproc) .gt. 0) then
          s_count(irnk) = 0
       end if
    end do

    call tpf_begin(tpf_mpi)
    call MPI_ALLTOALLV(tdp,r_count,r_displ,MPI_DOUBLE_COMPLEX, &
         dvxp,s_count,s_displ,MPI_DOUBLE_COMPLEX,comm,ierr)
    call tpf_end(tpf_mpi)

    deallocate(tdp,tdvxp,bi_count,bi_displ,bj_count,bj_displ, &
         s_count,s_displ,r_count,r_displ)

    call tpf_end(tpf_potexch)

    return
  end subroutine potexchg

  subroutine engexch_subr(va,bb,mgsf,nks, &
       nw,nb1,nb2,wgtk,bzwk,dpf,bkmi,ekmi,iks, &
       dp,pexeng,pexengs,nproc,g_count,g_displ,comm,mssg)
    use types
    use symwfn
    use timeprof
    use mpi
    use meshgrid
    implicit none

    real(kind=8),intent(in):: va,bb(3,3)
    type (meshgrid_struct):: mgsf
    integer,intent(in):: nks
    integer,intent(in):: nw
    integer,intent(in):: mssg
    type (wfn_for_k):: dpf(nks)
    integer,intent(in):: bkmi,ekmi,iks
    complex(kind=8):: dp(bkmi:ekmi,nw)
    real(kind=8):: pexeng,pexengs(6)
    integer,intent(in):: nb1,nb2
    real(kind=8),intent(in):: wgtk(nb1:nb2,nks),bzwk(nks)
    integer,intent(in):: nproc,g_count(0:nproc-1),g_displ(0:nproc-1),comm

    integer:: ng,nnp
    integer,pointer:: rg(:,:,:),pg(:,:),srg(:)
    integer:: ldkg
    integer,pointer:: nkmary(:),kg(:,:,:)
    integer:: nkmesh,nwx2,nwy2,nwz2

    real(kind=8):: skj(3),trssk(3),sq(3)
    integer,allocatable:: trskg(:,:)
    real(kind=8),allocatable:: pxi(:,:,:,:,:),pxj(:,:,:,:,:)
    complex(kind=8):: ctemp
    real(kind=8),allocatable:: wgti(:),wgtj(:)
    complex(kind=8),allocatable:: dpfwf(:,:),rbufwf(:)

    integer:: nlx,nly,nlz,nlxyz
    integer:: ix,iy,iz
    integer:: miw, mjw

    real(kind=8),parameter:: pi = 3.1415926535897932385d0

    integer:: iib,ijb,ib,jb,ig,i1,i2,i3,jks,irg,jrg
    real(kind=8):: ph
    complex(kind=8),allocatable:: zph(:)
    integer:: iph
    integer:: nb2eff
    real(kind=8):: epsnb2 = 64.0d0*epsilon(1.0d0)

    complex(kind=8),dimension(:),allocatable:: tdp
    integer,dimension(:),allocatable:: bi_count,bi_displ,bj_count,bj_displ
    integer,dimension(:),allocatable:: s_count,s_displ,r_count,r_displ
    integer:: myrank,irnk,krnk,iig,jg,ic,bjw,ejw

    integer:: ierr

    external setup_exchange_clkern, setup_exchange_clkerns, compute_excheng

    call tpf_begin(tpf_potexch)

    ng = mgsf%ng
    nnp = mgsf%nnp
    rg => mgsf%rg
    pg => mgsf%pg
    srg => mgsf%srg
    ldkg = mgsf%ldkg
    nkmary => mgsf%nkm
    kg => mgsf%kg
    nkmesh = mgsf%nkmesh
    nwx2 = mgsf%nwx*2
    nwy2 = mgsf%nwy*2
    nwz2 = mgsf%nwz*2

    call get_exchange_nlxyz(nlx,nly,nlz)
    nlxyz = nlx*nly*nlz

    call MPI_COMM_RANK(comm,myrank,ierr)
    allocate(zph(-nnp+1:nnp-1),s_count(0:nproc-1),s_displ(0:nproc-1), &
         r_count(0:nproc-1),r_displ(0:nproc-1), stat=ierr)
    call chkalloc('engexch_subr','tmp',ierr)

    do ib=nb2,nb1,-1
       if ( abs(wgtk(ib,iks)) .gt. epsnb2 ) then
          exit
       end if
    end do
    nb2eff = ib

    call reflesh_commij(myrank,nproc,comm,mssg)

    allocate(bi_count(0:niproc-1),bi_displ(0:niproc-1), &
         bj_count(0:njproc-1),bj_displ(0:njproc-1))

    bi_count = nw/niproc
    do irnk=0,mod(nw,niproc)-1
       bi_count(irnk) = bi_count(irnk) + 1
    end do
    bi_displ(0) = 0
    do irnk=0,niproc-2
       bi_displ(irnk+1) = bi_displ(irnk) + bi_count(irnk)
    end do

    allocate(FFT_A3M_DEFSZ(pxi,nlx,nly,nlz,bi_count(myirank)))
    allocate(tdp(nkmary(iks)*bi_count(myirank)),stat=ierr)
    call chkalloc('engexch_subr','tmp2',ierr)

    do krnk=0,nproc-1
       irnk = mod(krnk,niproc)
       s_count(krnk) = g_count(myrank)*bi_count(irnk)
       s_displ(krnk) = g_count(myrank)*bi_displ(irnk)
    end do

    call tpf_begin(tpf_mpi)
    call MPI_ALLTOALL(s_count,1,MPI_INTEGER,r_count,1,MPI_INTEGER,comm,ierr)
    call tpf_end(tpf_mpi)
    r_displ(0) = 0
    do irnk=0,nproc-2
       r_displ(irnk+1) = r_displ(irnk) + r_count(irnk)
    end do

    call tpf_begin(tpf_mpi)
    call MPI_ALLTOALLV(dp,s_count,s_displ,MPI_DOUBLE_COMPLEX, &
         tdp,r_count,r_displ,MPI_DOUBLE_COMPLEX,comm,ierr)
    call tpf_end(tpf_mpi)
    
    call clear_tmp(pxi,nlxyz*2*bi_count(myirank))
!$omp parallel do private(ig,ix,iy,iz,i1,i2,i3,irnk,ib,jg,iig)
    do irnk=0,nproc-1
       do ib=1,bi_count(myirank)
          jg = r_displ(irnk)+g_count(irnk)*(ib-1)
          iig = g_displ(irnk)
!OCL NOVREC(pxi)
!VOCL LOOP,DISJOINT(pxi)
!CDIR NODEP(pxi)
!DEC$ IVDEP
!VOPTION INDEP(pxi)
          do ig=1,g_count(irnk)
             ix = kg(1,iig+ig,iks)
             iy = kg(2,iig+ig,iks)
             iz = kg(3,iig+ig,iks)
             i1=mod(nwx2+ix,nwx2)+1
             i2=mod(nwy2+iy,nwy2)+1
             i3=mod(nwz2+iz,nwz2)+1

             FFT_A3M_RL(pxi,i1,i2,i3,ib) = dble(tdp(jg+ig))
             FFT_A3M_IM(pxi,i1,i2,i3,ib) = aimag(tdp(jg+ig))
          end do
       end do
    end do
    deallocate(tdp)

    allocate(trskg(3,ldkg), stat=ierr)
    allocate(FFT_A3M_DEFSZ(pxj,nlx,nly,nlz,blk_nw), &
         wgti(blk_nw),wgtj(blk_nw))

    do iph=-nnp+1,nnp-1
       ph = 2.0d0*pi/dble(nnp)*iph
       zph(iph) = dcmplx( cos(ph),-sin(ph) )
    end do

    do jks = 1, nks
       do jb=nb2,nb1,-1
          if ( abs(wgtk(jb,jks)) .gt. epsnb2 ) then
             exit
          end if
       end do
       nb2eff = jb

       bj_count = nb2eff/njproc
       do irnk=0,mod(nb2eff,njproc)-1
          bj_count(irnk) = bj_count(irnk) + 1
       end do
       bj_displ(0) = 0
       do irnk=0,njproc-2
          bj_displ(irnk+1) = bj_displ(irnk) + bj_count(irnk)
       end do

       allocate(dpfwf(nkmary(jks),bj_count(myjrank)))
       allocate(rbufwf(nkmary(jks)*bj_count(myjrank)))

       call tpf_begin(tpf_mpi)

       do krnk=0,nproc-1
          irnk = krnk/niproc
          s_count(krnk) = mgsf%g_count(myrank,jks)*bj_count(irnk)
          s_displ(krnk) = mgsf%g_count(myrank,jks)*bj_displ(irnk)
       end do

       call MPI_ALLTOALL(s_count,1,MPI_INTEGER,r_count,1,MPI_INTEGER,comm,ierr)
       r_displ(0) = 0
       do irnk=0,nproc-2
          r_displ(irnk+1) = r_displ(irnk) + r_count(irnk)
       end do
       call MPI_ALLTOALLV(dpf(jks)%wf,s_count,s_displ,MPI_DOUBLE_COMPLEX, &
            rbufwf,r_count,r_displ,MPI_DOUBLE_COMPLEX,comm,ierr)

       call tpf_end(tpf_mpi)

!$omp parallel do private(irnk,ic,jb,jg,ig)
       do irnk=0,nproc-1
          ic = r_displ(irnk)
          jg = mgsf%g_displ(irnk,jks)
          do jb=1,bj_count(myjrank)
             do ig=1,mgsf%g_count(irnk,jks)
                ic = ic + 1
                dpfwf(jg+ig,jb) = rbufwf(ic)
             end do
          end do
       end do

       deallocate(rbufwf)

       skj = mgsf%irdkpt(jks)%sk
       do irg=1,mgsf%irdkpt(jks)%ngir
          jrg = mgsf%irdkpt(jks)%irrg(irg)
          call transskkg(skj,nkmary(jks),kg(:,:,jks), &
               rg(:,:,jrg),srg(jrg),trssk,trskg)

          sq = mgsf%irdkpt(iks)%sk - trssk

          call tpf_begin(tpf_exclkern)
          call setup_exchange_clkern(nkmesh,bb,sq,rc,rd)
          call setup_exchange_clkerns(nkmesh,bb,sq,rc,rd)
          call tpf_end(tpf_exclkern)

          bjw = bj_displ(myjrank)+1
          ejw = bj_displ(myjrank)+bj_count(myjrank)
          do ijb=bjw,ejw,blk_nw
             mjw = min(ijb+blk_nw-1,ejw) - ijb + 1

             do jb=ijb,min(ijb+blk_nw-1,nb2eff)
                if (nb1.le.jb .and. jb.le.nb2eff) then
                   !
                   ! spin degeneracy is included in bzwk(jks)
                   !
                   wgtj(jb-ijb+1) = wgtk(jb,jks)/bzwk(jks)/dble(nkmesh)/va
                else
                   wgtj(jb-ijb+1) = 1.0d0/dble(nkmesh)/va
                end if
             end do

             call clear_tmp(pxj,nlxyz*2*blk_nw)
             if (srg(mgsf%irdkpt(jks)%irrg(irg)).eq.1) then
!$omp parallel do private(ig,ctemp,iph,ix,iy,iz,i1,i2,i3)
                do jb=ijb,min(ijb+blk_nw-1,ejw)
                   do ig=1,nkmary(jks)
                      ctemp = dpfwf(ig,jb-bj_displ(myjrank))
                      iph = pg(1,jrg)*trskg(1,ig)+pg(2,jrg)*trskg(2,ig) &
                           +pg(3,jrg)*trskg(3,ig)
                      iph = mod(iph,nnp)
                      ctemp = zph(iph)*ctemp

                      ix = trskg(1,ig)
                      iy = trskg(2,ig)
                      iz = trskg(3,ig)
                      i1=mod(nwx2+ix,nwx2)+1
                      i2=mod(nwy2+iy,nwy2)+1
                      i3=mod(nwz2+iz,nwz2)+1
                      FFT_A3M_RL(pxj,i1,i2,i3,jb-ijb+1) = dble(ctemp)
                      FFT_A3M_IM(pxj,i1,i2,i3,jb-ijb+1) = aimag(ctemp)
                   end do
                end do
             else
!$omp parallel do private(ig,ctemp,iph,ix,iy,iz,i1,i2,i3)
                do jb=ijb,min(ijb+blk_nw-1,ejw)
                   do ig=1,nkmary(jks)
                      ctemp = dpfwf(ig,jb-bj_displ(myjrank))
                      iph = pg(1,jrg)*trskg(1,ig)+pg(2,jrg)*trskg(2,ig) &
                           +pg(3,jrg)*trskg(3,ig)
                      iph = mod(iph,nnp)
                      ctemp = zph(iph)*conjg(ctemp)

                      ix = trskg(1,ig)
                      iy = trskg(2,ig)
                      iz = trskg(3,ig)
                      i1=mod(nwx2+ix,nwx2)+1
                      i2=mod(nwy2+iy,nwy2)+1
                      i3=mod(nwz2+iz,nwz2)+1
                      FFT_A3M_RL(pxj,i1,i2,i3,jb-ijb+1) = dble(ctemp)
                      FFT_A3M_IM(pxj,i1,i2,i3,jb-ijb+1) = aimag(ctemp)
                   end do
                end do
             end if

             !
             ! deploy_pxj_kern() may modify pxj itself
             !
             call deploy_pxj_kern(mjw,pxj)

             do iib=1,bi_count(myirank),blk_nw
                miw = min( iib+blk_nw-1, bi_count(myirank) ) - iib + 1

                do ib=iib,min(iib+blk_nw-1,bi_count(myirank))
                   if (bi_displ(myirank)+ib.lt.nb1) then
                      wgti(ib-iib+1) = bzwk(iks)
                   else if (bi_displ(myirank)+ib.le.nb2) then
                      wgti(ib-iib+1) = wgtk(bi_displ(myirank)+ib,iks)
                   end if
                end do

                call tpf_begin(tpf_cexch)
                call compute_excheng(miw,mjw, &
                     FFT_A3M_RL(pxi,1,1,1,iib),pxj, &
                     pexeng,pexengs,wgti,wgtj)
                call tpf_end(tpf_cexch)
             end do
          end do
       end do

       deallocate(dpfwf)
    end do
    deallocate(pxi,trskg,pxj,wgti,wgtj)

    deallocate(bi_count,bi_displ,bj_count,bj_displ, &
         s_count,s_displ,r_count,r_displ,zph)

    call tpf_end(tpf_potexch)

    return
  end subroutine engexch_subr

  subroutine engexch(va,bb,mgs,nspin,nks, &
       nw,nb1,nb2,wgtk,bzwk,fac, &
       dp,exeng,exengs,nproc,comm,mssg)
    use types
    use timeprof
    use mpi
    use meshgrid
    implicit none

    real(kind=8),intent(in):: va,bb(3,3)
    type (meshgrid_struct):: mgs
    integer,intent(in):: nks,nspin
    integer,intent(in):: nw
    integer,intent(in):: mssg
    type (wfn_for_k):: dp(nks,nspin)
    integer,intent(in):: nb1,nb2
    real(kind=8),intent(in):: wgtk(nb1:nb2,nks,nspin),bzwk(nks),fac
    real(kind=8):: exeng,exengs(6)
    integer,intent(in):: nproc,comm

    real(kind=8):: texeng,pexeng,texengs(6),pexengs(6)
    integer:: bkmi,ekmi
    integer:: is,iks
    integer:: myrank,ierr
    
    call MPI_COMM_RANK(comm,myrank,ierr)

    pexeng = 0.0d0
    pexengs = 0.0d0
    do is=1,nspin
       do iks=1,nks
          call engexch_subr(va,bb,mgs,nks, &
               nw,nb1,nb2,wgtk(:,:,is),bzwk, &
               dp(:,is),bkmi,ekmi,iks, &
               dp(iks,is)%wf,pexeng,pexengs, &
               nproc,mgs%g_count(:,iks),mgs%g_displ(:,iks),comm,mssg)
       end do
    end do

    call tpf_begin(tpf_mpi)
    call MPI_ALLREDUCE(pexeng,texeng,1,MPI_DOUBLE_PRECISION,MPI_SUM,comm,ierr)
    call MPI_ALLREDUCE(pexengs,texengs,6,MPI_DOUBLE_PRECISION,MPI_SUM,comm,ierr)
    call tpf_end(tpf_mpi)

    exeng = fac*texeng/mgs%nkmesh
    exengs = fac*texengs/mgs%nkmesh
    exengs(1:3) = exeng + exengs(1:3)

    return
  end subroutine engexch

  subroutine dm2ftm(nrx2,nry2,nrz2,nlx,nly,nlz,dm,ftm)
    implicit none
    integer,intent(in):: nrx2,nry2,nrz2,nlx,nly,nlz
    real(kind=8),intent(in):: dm(nrx2,nry2,nrz2)
    real(kind=8):: FFT_A3D_DEFSZ(ftm,nlx,nly,nlz)
    integer:: i1,i2,i3
    do i3=1,nrz2
       do i2=1,nry2
          do i1=1,nrx2
             FFT_A3D_RL(ftm,i1,i2,i3) = dm(i1,i2,i3)
             FFT_A3D_IM(ftm,i1,i2,i3) = 0.0d0
          end do
          do i1=nrx2,nlx
             FFT_A3D_RL(ftm,i1,i2,i3) = 0.0d0
             FFT_A3D_IM(ftm,i1,i2,i3) = 0.0d0
          end do
       end do
       do i2=nry2,nly
          FFT_A3D_RL(ftm,:,i2,i3) = 0.0d0
          FFT_A3D_IM(ftm,:,i2,i3) = 0.0d0
       end do
    end do
    do i3=nrz2,nlz
       FFT_A3D_RL(ftm,:,:,i3) = 0.0d0
       FFT_A3D_IM(ftm,:,:,i3) = 0.0d0
    end do
    return
  end subroutine dm2ftm

  subroutine pothybrid(va,bb,nspin,ndens,spaxs,brhoin,lpt,bexa,bexc &
       ,bevxc,nkfxa,kf,nrx2,nry2,nrz2,nvx,nvy,nvz,hyblpt,xctp,mssg,comm)
    use fft_3d
    use xcf_types
    use mpi
    use potihxc
    implicit none
    real(kind=8),intent(in):: va,bb(3,3)
    integer,intent(in):: nspin,ndens
    real(kind=8),intent(in):: spaxs(3)
    integer,intent(in):: nkfxa,kf(3,nkfxa)
    integer,intent(in):: nrx2,nry2,nrz2,nvx,nvy,nvz
    integer,intent(in):: mssg,comm
    integer,intent(in):: xctp
    real(kind=8),intent(out):: bexa,bexc,bevxc(0:6)
    real(kind=8),intent(in):: brhoin(nrx2,nry2,nrz2,nspin*ndens)
    complex(kind=8),intent(in):: lpt(-nvx:nvx,-nvy:nvy,-nvz:nvz,nspin*ndens)
    complex(kind=8):: hyblpt(-nvx:nvx,-nvy:nvy,-nvz:nvz,nspin*ndens)
    !
    integer:: nlx,nly,nlz,nlxyz
    real(kind=8),allocatable:: brho(:,:),rhoc(:),ztmp(:,:),wtmp(:,:)
    real(kind=8),allocatable:: tvxc4(:,:)

    integer:: ierr
    integer:: is,bxctp
    type (fft3_struct):: fs
    integer:: mspin
    integer:: nrxyz

    hyblpt = lpt
    if (xcf_exch_factor(xctp) .eq. 0.0d0) then
       bexa = 0.0d0
       bexc = 0.0d0
       bevxc = 0.0d0
       return
    end if

    if (ndens.eq.1) then
       mspin = nspin
    else
       mspin = 2
    end if

    nlx = FFT_PAD_NX(nrx2)
    nly = FFT_PAD_NY(nry2)
    nlz = FFT_PAD_NZ(nrz2)

    nlxyz=nlx*nly*nlz
    call fft3_init(nrx2,nry2,nrz2,nlx,nly,nlz,fs)

    nrxyz = nrx2*nry2*nrz2
    allocate(rhoc(nrxyz),brho(nlxyz,nspin),stat=ierr)
    call chkalloc('pothybrid','rhoc,brho',ierr)

    if (ndens.eq.1) then
       do is=1,nspin
          call rhovin2v(nrx2,nry2,nrz2,nlx,nly,nlz, &
               brhoin(1,1,1,is),brho(1,is))
       end do
    else
       call rho4to2c(nrxyz,spaxs,brhoin,brho)
    end if
    !
    ! partial core is zero
    !
    rhoc = 0.0d0

    select case (xctp)
    case (xcf_pbe0)
       bxctp = xcf_pbe0_x
       if (mssg.gt.0) write(6,*) 'pothybrid: compensation with PBE0_X'
    case (xcf_hartree_fock)
       bxctp = xcf_pbe
       if (mssg.gt.0) write(6,*) 'pothybrid: compensation with PBE'
    case (xcf_b3lyp)
       bxctp = xcf_b3lyp_x
       if (mssg.gt.0) write(6,*) 'pothybrid: compensation with B3LYP_X'
    case default
       if (mssg.gt.-1) write(6,*) 'invalid xctp in pothybrid()'
       call MPI_FINALIZE(ierr)
       stop
    end select

    call potxc(va,bb,mspin,fs,rhoc,brho,bexa,bexc &
       ,bevxc,nkfxa,kf,nrx2,nry2,nrz2,nlx,nly,nlz,bxctp,-1,comm)

    allocate(FFT_A1D_DEFSZ(ztmp,nlxyz),FFT_A1D_DEFSZ(wtmp,nlxyz), &
         stat=ierr)
    call chkalloc('pothybrid','ztmp,wtmp',ierr)
    ztmp = 0.0d0
    if (ndens.eq.1) then
       do is=1,nspin
          FFT_A1D_RL(ztmp,:) = brho(:,is)
          FFT_A1D_IM(ztmp,:) = 0.0d0
          call fft3_fw(fs,ztmp,wtmp)

          call sublpt(nlx,nly,nlz,nrx2,nry2,nrz2,nvx,nvy,nvz, &
               nkfxa,kf,hyblpt(:,:,:,is),ztmp)
       end do
    else
       allocate(tvxc4(nrxyz,4))
       call vxc2cto4(nrxyz,spaxs,brho,brhoin,tvxc4)
       do is=1,4
          call dm2ftm(nrx2,nry2,nrz2,nlx,nly,nlz,tvxc4(:,1),ztmp)
          call fft3_fw(fs,ztmp,wtmp)
          call sublpt(nlx,nly,nlz,nrx2,nry2,nrz2,nvx,nvy,nvz, &
               nkfxa,kf,hyblpt(:,:,:,is),ztmp)
       end do
       deallocate(tvxc4)
    end if

    deallocate(rhoc,ztmp,wtmp,brho)
    call fft3_done(fs)

    return
  end subroutine pothybrid

  subroutine sublpt(nlx,nly,nlz,nrx2,nry2,nrz2,nvx,nvy,nvz, &
       nkfxa,kf,potx,poty)
    implicit none
    integer,intent(in):: nlx,nly,nlz,nrx2,nry2,nrz2,nkfxa,kf(3,nkfxa)
    integer,intent(in):: nvx,nvy,nvz
    complex(kind=8):: potx(-nvx:nvx,-nvy:nvy,-nvz:nvz)
    real(kind=8):: FFT_A3D_DEFSZ(poty,nlx,nly,nlz)
    integer:: k1,k2,k3,l1,l2,l3,k

!$omp parallel do private(k1,k2,k3,l1,l2,l3)
!OCL NOVREC(potx)
!CDIR NODEP(potx)
!VOPTION INDEP(potx)
!POPTION INDEP(potx)
    do k=1,nkfxa
       k1=kf(1,k)
       k2=kf(2,k)
       k3=kf(3,k)
       l1=mod(nrx2+k1,nrx2)+1
       l2=mod(nry2+k2,nry2)+1
       l3=mod(nrz2+k3,nrz2)+1
       potx(k1,k2,k3) = potx(k1,k2,k3) & 
            - dcmplx(FFT_A3D_RL(poty,l1,l2,l3),FFT_A3D_IM(poty,l1,l2,l3))
    end do
    return
  end subroutine sublpt
end module exchange
